"""Pytest file for calling integration bash scripts."""

import os
import shlex
import subprocess
import sys

import pytest

# base, full, extra, and system marker test lists should be disjoint sets.
# pytest -m "base and full and extra and system"
# should result in a complete set of tests of the entire GeoIPS ecosystem, with
# no test duplication.

# Quick test to ensure basic functionality in geoips installation.
# The base marker should NOT require/test anything in any other plugin package.
base_integ_test_calls = [
    "$geoips_repopath/tests/scripts/amsr2.config_based_no_compare.sh",
    "$geoips_repopath/tests/scripts/amsr2_ocean.tc.windspeed.imagery_clean.sh",
]

# This includes ALL test scripts within the geoips repo itself with available data.
# The full marker should NOT require/test anything in any other plugin package.
full_integ_test_calls = [
    "$geoips_repopath/tests/utils/check_code.sh all $geoips_repopath",
    "$geoips_repopath/docs/build_docs.sh $geoips_repopath $geoips_pkgname html_only",
    "$geoips_repopath/tests/scripts/abi.static.Infrared.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/abi.static.Infrared.imagery_annotated_enhanced.sh",
    "$geoips_repopath/tests/scripts/console_script_create_sector_image.sh",
    "$geoips_repopath/tests/scripts/abi.static.Visible.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/abi.static.nasa_dust_rgb.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/abi.config_based_output_low_memory.sh",
    "$geoips_repopath/tests/scripts/abi.config_based_output.sh",
    "$geoips_repopath/tests/scripts/ami.static.Infrared.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/ami.static.Visible.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/ami.static.mst.absdiff-IR-BD.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/ami.tc.WV.geotiff.sh",
    "$geoips_repopath/tests/scripts/ami.WV-Upper.unprojected_image.sh",
    "$geoips_repopath/tests/scripts/amsr2.global.89H-Physical.cogeotiff.sh",
    "$geoips_repopath/tests/scripts/amsr2.tc.89H-Physical.imagery_annotated.sh",
    "$geoips_repopath/tests/scripts/amsr2_rss.tc.windspeed.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/amsr2.config_based_overlay_output.sh",
    "$geoips_repopath/tests/scripts/amsr2.config_based_overlay_output_low_memory.sh",
    "$geoips_repopath/tests/scripts/ascat_knmi.tc.windbarbs.imagery_windbarbs_clean.sh",
    "$geoips_repopath/tests/scripts/ascat_low_knmi.tc.windbarbs.imagery_windbarbs.sh",
    "$geoips_repopath/tests/scripts/ascat_noaa_25km.tc.windbarbs.imagery_windbarbs.sh",
    (
        "$geoips_repopath/tests/scripts/"
        "ascat_noaa_50km.tc.wind-ambiguities.imagery_windbarbs.sh"
    ),
    "$geoips_repopath/tests/scripts/ascat_uhr.tc.wind-ambiguities.imagery_windbarbs.sh",
    "$geoips_repopath/tests/scripts/ascat_uhr.tc.nrcs.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/ascat_uhr.tc.windbarbs.imagery_windbarbs.sh",
    "$geoips_repopath/tests/scripts/ascat_uhr.tc.windspeed.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/cli_dummy_script.sh",
    "$geoips_repopath/tests/scripts/gmi.tc.89pct.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/imerg.tc.Rain.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/oscat_knmi.tc.windbarbs.imagery_windbarbs.sh",
    "$geoips_repopath/tests/scripts/sar.tc.nrcs.imagery_annotated.sh",
    "$geoips_repopath/tests/scripts/smap.unsectored.text_winds.sh",
    "$geoips_repopath/tests/scripts/viirsday.global.Night-Vis-IR.cogeotiff_rgba.sh",
    "$geoips_repopath/tests/scripts/viirsday.tc.Night-Vis-IR.imagery_annotated.sh",
    "$geoips_repopath/tests/scripts/viirsmoon.tc.Night-Vis-GeoIPS1.imagery_clean.sh",
    (
        "$geoips_repopath/tests/scripts/"
        "seviri.WV-Upper.no_self_register.unprojected_image.sh"
    ),
    (
        "$geoips_repopath/tests/scripts/"
        "viirsclearnight.Night-Vis-IR-GeoIPS1.unprojected_image.sh"
    ),
    "$geoips_repopath/tests/scripts/smap.awips.netcdf.sh",
    "$geoips_repopath/tests/scripts/smos.awips.netcdf.sh",
]

# Test scripts spanning multiple repositories / geoips plugins.
# The system marker allows us to test the full GeoIPS open source system implementation.
# Note system does NOT test the current geoips repo, as we can assume we will test the
# current repo through the base and/or full test.
system_integ_test_calls = [
    # Include this in system test, since it runs across all plugin repos.
    "python $GEOIPS_PACKAGES_DIR/geoips/tests/utils/test_interfaces.py",
    # Include this in system test, since it runs across all plugin repos.
    "$geoips_repopath/tests/scripts/console_script_list_available_plugins.sh",
]

# Test scripts that require extra datasets.
extra_integration_test_calls = [
    "$geoips_repopath/tests/scripts/abi.config_based_dmw_overlay.sh",
    "$geoips_repopath/tests/scripts/abi.static.dmw.imagery_windbarbs_high.sh",
    "$geoips_repopath/tests/scripts/amsua_mhs_mirs.tc.183-3H.imagery_annotated.sh",
    "$geoips_repopath/tests/scripts/atms.tc.165H.netcdf_geoips.sh",
    "$geoips_repopath/tests/scripts/ewsg.static.Infrared.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/fci.static.Visible.imagery_annotated.sh",
    "$geoips_repopath/tests/scripts/fci.unprojected_image.Infrared.sh",
    "$geoips_repopath/tests/scripts/hy2.tc.windspeed.imagery_annotated.sh",
    "$geoips_repopath/tests/scripts/mimic_coarse.static.TPW-CIMSS.imagery_annotated.sh",
    "$geoips_repopath/tests/scripts/mimic_fine.tc.TPW-PWAT.imagery_annotated.sh",
    "$geoips_repopath/tests/scripts/modis.Infrared.unprojected_image.sh",
    "$geoips_repopath/tests/scripts/saphir.tc.183-3HNearest.imagery_annotated.sh",
    "$geoips_repopath/tests/scripts/seviri.WV-Upper.unprojected_image.sh",
    "$geoips_repopath/tests/scripts/seviri.airmass.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/seviri.Convective_Storms.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/seviri.Day_Microphys_Summer.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/seviri.Day_Microphys_Winter.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/seviri.Day_Solar.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/seviri.Dust_RGB.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/seviri.Natural_Color.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/seviri.Night_Microphys.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/seviri.Volcanic_Ash.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/smap.unsectored.text_winds.sh",
    "$geoips_repopath/tests/scripts/smos.tc.sectored.text_winds.sh",
    "$geoips_repopath/tests/scripts/ssmi.tc.37pct.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/ssmis.color91.unprojected_image.sh",
    "$geoips_repopath/tests/scripts/gfs.static.windspeed.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/gfs.static.waveheight.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/viirs.static.visible.imagery_clean.sh",
]


def check_base_install():
    """
    Run the base installation check script to verify the GeoIPS installation.

    Executes the 'base_install.sh' script located in the GeoIPS tests directory
    to ensure that all required components are properly installed.

    Raises
    ------
    subprocess.CalledProcessError
        If the installation check script returns a non-zero exit status.
    """
    base_install_check = [
        "bash",
        os.path.join(
            os.getenv("GEOIPS_PACKAGES_DIR"),
            "geoips/tests/integration_tests/base_install.sh",
        ),
        "exit_on_missing",
    ]
    print("Running base install check")
    subprocess.check_output(base_install_check, env=os.environ.copy())


def check_full_install():
    """
    Run the full installation check script to verify the GeoIPS installation.

    Executes the 'full_install.sh' script located in the GeoIPS tests directory
    to ensure that all required components are properly installed.

    Raises
    ------
    subprocess.CalledProcessError
        If the installation check script returns a non-zero exit status.
    """
    full_install_check = [
        "bash",
        os.path.join(
            os.getenv("GEOIPS_PACKAGES_DIR"),
            "geoips/tests/integration_tests/full_install.sh",
        ),
        "exit_on_missing",
    ]
    print("Running full install check")
    subprocess.check_output(full_install_check, env=os.environ.copy())


def check_system_install():
    """
    Run the system installation check script to verify the GeoIPS installation.

    Executes the 'system_install.sh' script located in the GeoIPS tests directory
    to ensure that all required components are properly installed.

    Raises
    ------
    subprocess.CalledProcessError
        If the installation check script returns a non-zero exit status.
    """
    system_install_check = [
        "bash",
        os.path.join(
            os.getenv("GEOIPS_PACKAGES_DIR"),
            "geoips/tests/integration_tests/system_install.sh",
        ),
        "exit_on_missing",
    ]
    print("Running system install check")
    subprocess.check_output(system_install_check, env=os.environ.copy())


def setup_environment():
    """
    Set up necessary environment variables for integration tests.

    Configures paths and package names for the GeoIPS core and its plugins by
    setting environment variables required for the integration tests. Assumes
    that 'GEOIPS_PACKAGES_DIR' is already set in the environment.

    Notes
    -----
    The following environment variables are set:
    - geoips_repopath
    - geoips_pkgname
    """
    # Base path
    os.environ["geoips_repopath"] = os.path.join(os.path.dirname(__file__), "..", "..")
    os.environ["geoips_pkgname"] = "geoips"

    # Environment variable for GEOIPS_PACKAGES_DIR
    geoips_packages_dir = os.getenv("GEOIPS_PACKAGES_DIR")
    if not geoips_packages_dir:
        raise EnvironmentError("GEOIPS_PACKAGES_DIR environment variable not set.")


@pytest.fixture(scope="session")
def system_setup():
    """
    Set up the system integration tests by checking install and setting env-vars.

    Calls `check_system_install()` to verify the installation and `setup_environment()`
    to set up the necessary environment variables before running integration tests.
    """
    check_system_install()
    setup_environment()


@pytest.fixture(scope="session")
def base_setup():
    """
    Set up the base integration tests by checking install and setting env-vars.

    Calls `check_base_install()` to verify the installation and `setup_environment()`
    to set up the necessary environment variables before running integration tests.
    """
    check_base_install()
    setup_environment()


@pytest.fixture(scope="session")
def full_setup():
    """
    Set up the full integration tests by checking install and setting env-vars.

    Calls `check_full_install()` to verify the installation and `setup_environment()`
    to set up the necessary environment variables before running integration tests.
    """
    check_full_install()
    setup_environment()


def run_script_with_bash(script):
    """
    Run scripts by executing specified shell commands with bash.

    Parameters
    ----------
    script : str
        Shell command to executes. The command may
        contain environment variables which will be expanded before execution.

    Raises
    ------
    subprocess.CalledProcessError
        If the shell command returns a non-zero exit status.
    """
    if ".sh" in script:
        expanded_call = shlex.split("bash " + os.path.expandvars(script))
    else:
        expanded_call = shlex.split(os.path.expandvars(script))
    print("Running: ", script)
    print("Expanded call: ", expanded_call)
    try:
        subprocess.check_output(expanded_call, env=os.environ.copy())
    except subprocess.CalledProcessError as e:
        print(e.output.decode(sys.stdout.encoding))
        raise e


@pytest.mark.base
@pytest.mark.integration
@pytest.mark.parametrize("script", base_integ_test_calls)
def test_integ_base_test_script(base_setup: None, script: str):
    """
    Run integration test scripts by executing specified shell commands.

    Parameters
    ----------
    script : str
        Shell command to execute as part of the integration test. The command may
        contain environment variables which will be expanded before execution.

    Raises
    ------
    subprocess.CalledProcessError
        If the shell command returns a non-zero exit status.
    """
    run_script_with_bash(script)


@pytest.mark.system
@pytest.mark.integration
@pytest.mark.parametrize("script", system_integ_test_calls)
def test_integ_system_test_script(system_setup: None, script: str):
    """
    Run integration test scripts by executing specified shell commands.

    Parameters
    ----------
    script : str
        Shell command to execute as part of the integration test. The command may
        contain environment variables which will be expanded before execution.

    Raises
    ------
    subprocess.CalledProcessError
        If the shell command returns a non-zero exit status.
    """
    run_script_with_bash(script)


@pytest.mark.full
@pytest.mark.integration
@pytest.mark.parametrize("script", full_integ_test_calls)
def test_integ_full_test_script(full_setup: None, script: str):
    """
    Run integration test scripts by executing specified shell commands.

    Parameters
    ----------
    script : str
        Shell command to execute as part of the integration test. The command may
        contain environment variables which will be expanded before execution.

    Raises
    ------
    subprocess.CalledProcessError
        If the shell command returns a non-zero exit status.
    """
    run_script_with_bash(script)


@pytest.mark.extra
@pytest.mark.integration
@pytest.mark.parametrize("script", extra_integration_test_calls)
def test_integ_extra_test_script(full_setup: None, script: str):
    """
    Run integration test scripts by executing specified shell commands.

    Parameters
    ----------
    script : str
        Shell command to execute as part of the integration test. The command may
        contain environment variables which will be expanded before execution.

    Raises
    ------
    subprocess.CalledProcessError
        If the shell command returns a non-zero exit status.
    """
    run_script_with_bash(script)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""WFABBA ascii data reader.

WFABBA is a geostationary fire product produced by SSEC
"""

# Python Standard Libraries
from datetime import datetime
import logging
from os.path import basename

# Third-Party Libraries
import numpy as np
import xarray

LOG = logging.getLogger(__name__)

interface = "readers"
family = "standard"
name = "wfabba_ascii"
source_names = ["abi", "ahi", "gvar", "seviri"]


def parse_header_line(line):
    """Parse header line."""
    # Remove white space and '###' from header line
    line = line.replace("### ", "").strip()
    line = line.replace("   ", "")
    # There could be multiple meta data entries per line.
    # Entries are separated by commas
    if "," in line:
        entries = line.split(", ")
    else:
        entries = [line]
    # meta data field and description are separated by ':'
    # Loop through entries and split into key and value
    parsed_line = {}
    for entry in entries:
        if ":" not in entry:
            # Return None if entry does not follow key: val format
            return None
        split_entry = entry.split(": ")
        if len(split_entry) == 1:
            key = split_entry[0]
            val = ""
        else:
            key, val = split_entry
        if key == "Code":
            # Add special handling to keep information on the
            # different code meanings
            codes = [
                "10 (30) - Processed Fire Pixel (Temporally filtered)",
                "11 (31) - Saturated Fire Pixel (Temporally filtered)",
                "12 (32) - Cloudy Fire Pixel (Temporally filtered)",
                "13 (33) - High Probability Fire Pixel (Temporally filtered)",
                "14 (34) - Medium Probability Fire Pixel (Temporally filtered)",
                "15 (35) - Low Probability Fire Pixel (Temporally filtered)",
            ]
            val = "\n".join(codes)
        parsed_line[key] = val
    return parsed_line


def read_wfabba_header(wfabba_file):
    """Read WFABBA header."""
    with open(wfabba_file, "r") as f:
        is_header = True
        header_length = 0
        header_meta = {}
        while is_header:
            line = f.readline()
            if line[:3] == "###":
                parsed_line = parse_header_line(line)
                if parsed_line:
                    header_meta.update(parsed_line)
                header_length += 1
            else:
                is_header = False
    header_meta["header length"] = header_length - 2
    full_time = f"{header_meta.pop('Date')} {header_meta.pop('Time')}"
    header_meta["datetime"] = datetime.strptime(full_time, "%Y%j %H:%M:%S UTC")
    return header_meta


def read_wfabba_text(wfabba_file):
    """Read WFABBA text from file wfabba_file into xarray Dataset."""
    xobj = xarray.Dataset()
    header_meta = read_wfabba_header(wfabba_file)
    text = np.genfromtxt(
        wfabba_file,
        delimiter=",",
        skip_header=header_meta["header length"],
        comments=None,
        dtype=str,
    )
    header_meta["units"] = {}
    if int(header_meta["Number of detected fires"]) > 0:
        col_names = text[0, :]
        col_units = text[1, :]
        for col_num, (col_name, col_unit) in enumerate(zip(col_names, col_units)):
            col_name = col_name.replace("### ", "").replace(" ", "").lower()
            col_unit = col_unit.replace("### ", "").replace(" ", "")
            data = text[2:, col_num].astype(float)
            xobj[col_name] = xarray.DataArray(data)
            header_meta["units"][col_name] = col_unit
        xobj["firetime"] = xarray.DataArray(
            [header_meta["datetime"]] * int(header_meta["Number of detected fires"])
        )
    xobj.attrs["start_datetime"] = header_meta["datetime"]
    xobj.attrs["end_datetime"] = header_meta["datetime"]
    xobj.attrs["source_file_datetimes"] = [header_meta["datetime"]]
    xobj.attrs["platform_name"] = header_meta["Satellite"]
    xobj.attrs["source_name"] = header_meta["Instrument"]
    xobj.attrs["data_provider"] = header_meta["Data source"]
    return xobj


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read WFABBA ascii data from a list of filenames.

    WFABBA  ascii files contain list of fire detects with their latitude,
    longitude, and scan location

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * NOT YET IMPLEMENTED
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    xarray_objs = []
    metadata = read_wfabba_header(fnames[0])
    end_metadata = read_wfabba_header(fnames[-1])
    geoips_attrs = {
        "area_definition": area_def,
        "start_datetime": metadata["datetime"],
        "end_datetime": end_metadata["datetime"],
        "vertical_data_type": "surface",
        "source_name": "wfabba",
        "data_provider": "ssec",
        "interpolation_radius_of_influence": 2000,
    }
    if area_def:
        geoips_attrs["area_id"] = area_def.area_id
    else:
        geoips_attrs["area_id"] = metadata.get("scene_id")

    meta_dataset = xarray.Dataset(attrs=dict(metadata, **geoips_attrs))
    if metadata_only:
        return {"METADATA": meta_dataset}

    for i, fname in enumerate(fnames):
        LOG.info("Reading %s" % fname)
        wfabba_xobj = read_wfabba_text(fname)
        wfabba_xobj.attrs["source_file_names"] = [basename(fname)]
        wfabba_xobj.attrs["sample_distance_km"] = 2
        wfabba_xobj.attrs = dict(wfabba_xobj.attrs, **geoips_attrs)
        xarray_objs.append(wfabba_xobj)

    if len(xarray_objs) > 1:
        start_times = [x.attrs["start_datetime"] for x in xarray_objs]
        end_times = [x.attrs["end_datetime"] for x in xarray_objs]
        xarray_dset = xarray.concat(xarray_objs, dim="dim_0")
        xarray_dset.attrs["start_datetime"] = min(start_times)
        xarray_dset.attrs["end_datetime"] = max(end_times)
        xarray_dset = xarray_dset.assign_coords({"time": start_times})
    else:
        xarray_dset = xarray_objs[0]
    return {"wfabba": xarray_dset, "METADATA": meta_dataset}



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit test asserting functionality for exposing plugin-package commands."""

from importlib import metadata
import logging
import pytest

from geoips.errors import PluginPackageNotFoundError
from geoips.geoips_utils import expose_geoips_commands

LOG = logging.getLogger(__name__)

plugin_packages = [
    str(ep.value) for ep in metadata.entry_points(group="geoips.plugin_packages")
]


def generate_id(pkg_name):
    """Generate a test ID for the provided plugin package."""
    return f"expose {pkg_name}"


@pytest.mark.parametrize(
    "pkg_name", plugin_packages + ["non_existent_pkg"], ids=generate_id
)
def test_expose_pkg_cmds(pkg_name, caplog):
    """Assert various calls to 'expose_geoips_commands' work as expected.

    Parameters
    ----------
    pkg_name: str
        - The name of the plugin package whose command's that will be exposed
    caplog: pytest caplog fixture
        - Pytest fixture which captures the log from the corresponding calls
    """
    # 25 is the level of LOG.interactive, which we use in expose_geoips_commands
    caplog.set_level(25)
    eps = list(
        filter(
            lambda ep: pkg_name in ep.value,
            metadata.entry_points().select(group="console_scripts"),
        )
    )
    if len(eps):
        # Both of these packages have commands. geoips is implemented via poetry, and
        # data_fusion is implemented via setuptools. Either way, this should work
        # without an error being raised
        expose_geoips_commands(pkg_name, LOG)
        assert f"Available {pkg_name.title()} Commands" in caplog.text
        # Replace calls are needed as we need to filter out the table chars
        # if the table was split due to terminal size
        replaced = str(
            caplog.text.replace("\n", "")
            .replace(" ", "")
            .replace(
                "-",
                "",
            )
            .replace("│", "")
        )
        for ep in eps:
            assert ep.name in replaced  # name of the command
            assert ep.value in replaced  # path to the command
    elif pkg_name in plugin_packages:
        # None of these packaes have commands
        expose_geoips_commands(pkg_name, LOG)
        assert f"No '{pkg_name.title()}' Commands were found." in caplog.text
    else:
        # Calling this with a non-existent package should raise PackageNotFoundError
        with pytest.raises(PluginPackageNotFoundError):
            expose_geoips_commands(pkg_name, LOG)
    caplog.clear()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit test for GeoIPS CLI `list source-names` command.

See geoips/commandline/ancillary_info/cmd_instructions.yaml for more information.
"""

import json
from importlib import resources
from pathlib import Path
import pytest

from tests.unit_tests.commandline.cli_top_level_tester import BaseCliTest


class TestGeoipsListSourceNames(BaseCliTest):
    """Unit Testing Class for List Source Names Command."""

    @property
    def command_combinations(self):
        """A list of call signatures for the GeoipsListSourceNames command.

        This includes failing cases as well.
        """
        if not hasattr(self, "_cmd_list"):
            self._cmd_list = []
            base_args = ["geoips", "list", "source-names"]
            alias_args = ["geoips", "ls", "src-names"]
            for argset in [base_args, alias_args]:
                for pkg_name in self.plugin_package_names + ["all"]:
                    if pkg_name != "all":
                        args = argset + ["-p", pkg_name]
                    else:
                        args = argset
                    self._cmd_list.append(args)
            # Add argument list that utilizes the --column optional arg
            self._cmd_list.append(
                base_args + ["--columns", "source_name", "reader_names"]
            )
            self._cmd_list.append(
                alias_args + ["--columns", "source_name", "reader_names"]
            )
            # Add argument list with an existing interface but non-existent package
            self._cmd_list.append(base_args + ["-p", "non_existent_package"])
            self._cmd_list.append(alias_args + ["-p", "non_existent_package"])
            # Add argument list with an existing interface but w/ conflicting opt args
            self._cmd_list.append(
                base_args + ["--long", "--columns", "source_name", "reader_names"]
            )
            self._cmd_list.append(
                alias_args + ["--long", "--columns", "source_name", "reader_names"]
            )
        return self._cmd_list

    def check_error(self, args, error):
        """Check that the 'geoips list source-names ...' error output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI (expected to fail)
        error: str
            - Multiline str representing the error output of the CLI call
        """
        if "--columns" in args and "--long" in args:
            assert (
                "error: argument --columns/-c: not allowed with argument --long/-l"
                in error.replace("\n", "")
            )
        elif "-p" in args and args[-1] not in self.plugin_package_names:
            # check the package name is incorrect
            usg_str = (
                f"error: argument --package_name/-p: invalid "
                f"choice: '{args[-1]}' (choose from"
            )
            assert usg_str in error.replace("\n", "")
        assert (
            "usage: To use, type `geoips list source-names`" in error
            or "usage: To use, type `geoips list <cmd> <sub-cmd>`" in error
        )

    def check_output(self, args, output):
        """Check that the 'geoips list source-names...' success output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI
        output: str
            - Multiline str representing the output of the CLI call
        """
        # The args provided are valid, so test that the output is actually correct
        if "no reader plugins" in output and "-p" in args:
            # No plugins were found under the selected interface, within a
            # certain package; ensure that is correct.
            plugin_registry = json.load(
                open(
                    Path(resources.files(args[-1]).joinpath("registered_plugins.json")),
                    "r",
                ),
            )
            # assert that the provided interface doesn't exist within that package's
            # plugin registry
            assert "readers" not in plugin_registry["module_based"].keys()
        else:
            # Assert that the correct headers exist in the CLI output
            selected_cols = self.retrieve_selected_columns(args)
            headers = {
                "Source Name": "source_name",
                "Reader Names": "reader_names",
            }
            self.assert_correct_headers_in_output(output, headers, selected_cols)


test_sub_cmd = TestGeoipsListSourceNames()


@pytest.mark.parametrize(
    "args",
    test_sub_cmd.command_combinations,
    ids=test_sub_cmd.generate_id,
)
def test_command_combinations(monkeypatch, args):
    """Test all 'geoips list source-names ...' commands.

    This test covers valid combination of commands for the
    'geoips list source-names' command. We also test invalid commands, to ensure
    that the proper help documentation is provided for those using the command
    incorrectly.

    Parameters
    ----------
    args: 2D array of str
        - List of arguments to call the CLI with
          (ie. ['geoips', 'list', 'source-names'])
    """
    test_sub_cmd.test_command_combinations(monkeypatch, args)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Matplot-lib based annotated image output."""

import logging

from geoips.interfaces import products
from geoips.errors import PluginError
from jsonschema.exceptions import ValidationError

LOG = logging.getLogger(__name__)

interface = "output_formatters"
family = "image_overlay"
name = "imagery_annotated"


def call(
    area_def,
    xarray_obj,
    product_name,
    output_fnames,
    clean_fname=None,
    product_name_title=None,
    mpl_colors_info=None,
    feature_annotator=None,
    gridline_annotator=None,
    product_datatype_title=None,
    bg_data=None,
    bg_mpl_colors_info=None,
    bg_xarray=None,
    bg_product_name_title=None,
    bg_datatype_title=None,
    remove_duplicate_minrange=None,
    title_copyright=None,
    title_formatter=None,
    output_dict=None,
):
    """Plot annotated imagery."""
    if product_name_title is None:
        product_name_title = product_name

    success_outputs = []
    plot_data = xarray_obj[product_name].to_masked_array()
    from geoips.image_utils.mpl_utils import create_figure_and_main_ax_and_mapobj
    from geoips.image_utils.colormap_utils import set_matplotlib_colors_standard
    from geoips.image_utils.mpl_utils import (
        plot_image,
        save_image,
        plot_overlays,
        create_colorbar,
    )
    from geoips.image_utils.mpl_utils import get_title_string_from_objects, set_title

    bkgrnd_clr = None
    frame_clr = None
    # If a feature_annotator plugin was supplied, attempt to get the image background
    # color. Otherwise, just keep it as None.
    if feature_annotator:
        bkgrnd_clr = feature_annotator.get("spec", {}).get("background")
    # If a gridline_annotator plugin was supplied, attempt to get the frame background
    # color. Otherwise, just keep it as None.
    if gridline_annotator:
        frame_clr = gridline_annotator.get("spec", {}).get("background")

    if not mpl_colors_info:
        # Create the matplotlib color info dict - the fields in this dictionary
        # (cmap, norm, features, etc) will be used in plot_image to ensure the image
        # matches the colorbar.
        mpl_colors_info = set_matplotlib_colors_standard(
            data_range=[plot_data.min(), plot_data.max()],
            cmap_name=None,
            cbar_label=None,
        )
    mapobj = None
    if clean_fname:
        # Create matplotlib figure and main axis, where the main image will be plotted
        fig, main_ax, mapobj = create_figure_and_main_ax_and_mapobj(
            area_def.x_size,
            area_def.y_size,
            area_def,
            noborder=True,
            frame_clr=frame_clr,
        )

        # Plot the actual data on a map
        plot_image(
            main_ax,
            plot_data,
            mapobj,
            mpl_colors_info=mpl_colors_info,
            bkgrnd_clr=bkgrnd_clr,
        )

        LOG.info("Saving the clean image %s", clean_fname)
        # Save the clean image with no gridlines or coastlines
        success_outputs += save_image(
            fig,
            clean_fname,
            is_final=False,
            image_datetime=xarray_obj.start_datetime,
            remove_duplicate_minrange=remove_duplicate_minrange,
        )

    # Create matplotlib figure and main axis, where the main image will be plotted
    fig, main_ax, mapobj = create_figure_and_main_ax_and_mapobj(
        area_def.x_size,
        area_def.y_size,
        area_def,
        existing_mapobj=mapobj,
        noborder=False,
        frame_clr=frame_clr,
    )

    # Plot the actual data on a map
    plot_image(
        main_ax,
        plot_data,
        mapobj,
        mpl_colors_info=mpl_colors_info,
        bkgrnd_clr=bkgrnd_clr,
    )

    if bg_data is not None and (
        hasattr(plot_data, "mask") or len(plot_data.shape) == 3
    ):
        if not bg_mpl_colors_info:
            bg_mpl_colors_info = set_matplotlib_colors_standard(
                data_range=[plot_data.min(), plot_data.max()],
                cmap_name=None,
                cbar_label=None,
                create_colorbar=False,
            )
        import numpy

        # Plot the background data on a map. Support either RGBA arrays or masked arrays
        if len(plot_data.shape) == 3 and plot_data.shape[2] == 4:
            plot_image(
                main_ax,
                numpy.ma.masked_where(plot_data[:, :, 3], bg_data),
                mapobj,
                mpl_colors_info=bg_mpl_colors_info,
                bkgrnd_clr=bkgrnd_clr,
            )
        else:
            plot_image(
                main_ax,
                numpy.ma.masked_where(~plot_data.mask, bg_data),
                mapobj,
                mpl_colors_info=bg_mpl_colors_info,
                bkgrnd_clr=bkgrnd_clr,
            )

    # Set the title for final image
    title_string = get_title_string_from_objects(
        area_def,
        xarray_obj,
        product_name_title,
        product_datatype_title=product_datatype_title,
        bg_xarray=bg_xarray,
        bg_product_name_title=bg_product_name_title,
        bg_datatype_title=bg_datatype_title,
        title_copyright=title_copyright,
        title_formatter=title_formatter,
    )
    set_title(main_ax, title_string, area_def.y_size)

    if mpl_colors_info["colorbar"] is True:
        # Create the colorbar to match the mpl_colors
        create_colorbar(fig, mpl_colors_info)

    # Plot gridlines and feature overlays
    plot_overlays(
        mapobj,
        main_ax,
        area_def,
        feature_annotator=feature_annotator,
        gridline_annotator=gridline_annotator,
    )
    prod_plugin = None
    try:
        prod_plugin = products.get_plugin(
            xarray_obj.source_name,
            product_name,
            output_dict.get("product_spec_override"),
        )
    except (PluginError, ValidationError):
        LOG.warning(
            "SKIPPING products.get_plugin: Invalid product specification %s / %s",
            product_name,
            xarray_obj.source_name,
        )
    if prod_plugin and "coverage_checker" in prod_plugin:
        from geoips.dev.product import get_covg_from_product
        from importlib import import_module
        from geoips.dev.product import get_covg_args_from_product

        covg_func = get_covg_from_product(prod_plugin)
        covg_args = get_covg_args_from_product(prod_plugin)
        plot_covg_func = getattr(import_module(covg_func.__module__), "plot_coverage")
        plot_covg_func(main_ax, area_def, covg_args)

    if output_fnames is not None:
        for annotated_fname in output_fnames:
            # Save the final image
            success_outputs += save_image(
                fig,
                annotated_fname,
                is_final=True,
                image_datetime=xarray_obj.start_datetime,
                remove_duplicate_minrange=remove_duplicate_minrange,
            )

    return success_outputs



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""
The Geolocated Information Processing System (GeoIPS).

GeoIPS |unireg| Base Package
----------------------------

The GeoIPS Base Package provides a Python 3 based architecture supporting a wide
variety of satellite and weather data processing. The modular nature of the GeoIPS
base infrastructure also allows plug-and-play capability for user-specified custom
functionality.

Homepage: https://github.com/NRLMMD-GEOIPS/geoips

.. |unireg|    unicode:: U+000AE .. REGISTERED SIGN
"""
from matplotlib import rcParams
from geoips import errors
from geoips import filenames
from geoips import interfaces
from geoips import utils
from geoips import xarray_utils
from ._version import __version__, __version_tuple__

import logging  # noqa
from geoips.commandline.log_setup import add_logging_level

# Turn off image interpolation for matplotlib by default.
rcParams["image.interpolation"] = "none"

# Setting INTERACTIVE to 25 as this is between INFO (20) and WARNING (30). This way,
# WARNING output will still be shown if INFO or INTERACTIVE is set.
add_logging_level("INTERACTIVE", 25)

__all__ = [
    "interfaces",
    "errors",
    "filenames",
    "utils",
    "xarray_utils",
    "__version__",
    "__version_tuple__",
]



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Default TC metadata filename formatter."""

# metadata_yaml_filename = tc_fname_metadata(area_def, xarray_obj,
#                                            product_filename, metadata_dir,
#                                            metadata_type='sector_information',
#                                            basedir=basedir)

# Python Standard Libraries
from os.path import join as pathjoin
from os.path import basename
import logging

from geoips.filenames.base_paths import PATHS as gpaths
from geoips.plugins.modules.filename_formatters.utils.tc_file_naming import (
    tc_storm_basedir,
)

LOG = logging.getLogger(__name__)

interface = "filename_formatters"
family = "standard_metadata"
name = "metadata_default_fname"


def call(
    area_def,
    xarray_obj,
    product_filename,
    metadata_dir="metadata",
    metadata_type="sector_information",
    basedir=gpaths["TCWWW"],
    output_dict=None,
):
    """Generate TC metadata filenames.

    This uses attributes on both the xarray and area_def in order to produce
    the YAML metadata output specifically for TC sectors.  Not valid for
    other sector types.

    This uses the "tc_storm_basedir" utility to ensure a consistent path
    to the storm directory (so products and metadata end up in the same
    location)
    """
    from geoips.sector_utils.utils import is_sector_type

    if not is_sector_type(area_def, "tc"):
        return None

    allowed_metadata_types = [
        "sector_information",
        "archer_information",
        "storm_information",
    ]

    if metadata_type not in allowed_metadata_types:
        raise TypeError(
            f"Unknown metadata type {metadata_type}, allowed {allowed_metadata_types}"
        )

    tc_year = int(area_def.sector_info["storm_year"])
    tc_basin = area_def.sector_info["storm_basin"]
    tc_stormnum = int(area_def.sector_info["storm_num"])
    metadata_type = "sector_information"
    metadata_datetime = xarray_obj.start_datetime

    metadata_yaml_dirname = pathjoin(
        tc_storm_basedir(
            basedir,
            tc_year,
            tc_basin,
            tc_stormnum,
            output_dict=output_dict,
            sector_info=area_def.sector_info,
        ),
        metadata_dir,
        metadata_type,
        metadata_datetime.strftime("%Y%m%d"),
    )
    metadata_yaml_basename = basename(product_filename) + ".yaml"

    return pathjoin(metadata_yaml_dirname, metadata_yaml_basename)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Geotiff image rasterio-based output format."""

import os
import logging

import numpy
from affine import Affine
from rasterio.io import MemoryFile
from rasterio import uint8

# from rasterio.transform import from_bounds
from rasterio.plot import reshape_as_raster

# from rasterio.enums import ColorInterp
import rasterio
from rio_cogeo.cogeo import cog_translate
from rio_cogeo.profiles import cog_profiles

# from pyresample.utils import wrap_longitudes

# Internal utilities
from geoips.errors import (
    OutputFormatterInvalidProjectionError,
    OutputFormatterDatelineError,
)
from geoips.filenames.base_paths import make_dirs

LOG = logging.getLogger(__name__)

interface = "output_formatters"
family = "image"
name = "cogeotiff"


def get_rasterio_cmap_dict(
    mpl_cmap, scale_data_min=1, scale_data_max=255, missing_value=0
):
    """Get rasterio cmap dict, with vals between scale_data_min and scale_data_max."""
    # Scale from 1 to 255, leave 0 for missing value.
    cmap_arr = mpl_cmap(range(scale_data_min, scale_data_max)) * scale_data_max
    cmap_dict = {}
    arr_ind = 0
    # Get the actual RGBA values for each of the values in the colormap, with
    # a total number of colors between scale_data min and scale_data_max.
    for ii in range(scale_data_min, scale_data_max):
        cmap_dict[ii] = tuple(cmap_arr[arr_ind, :])
        arr_ind += 1
    # RGB values don't matter, just the "0" for the masked part.
    cmap_dict[missing_value] = (0, 255.0, 255.0, 0.0)
    return cmap_dict


def get_scales_and_offsets(
    bands, requested_data_min, requested_data_max, scale_data_min, scale_data_max
):
    """Get scale and offset information from dataset.

    These values are written to the geotiff file metadata in order to use for
    retrieving quantitative values from the scaled data when rendering.
    """
    scales = []
    offsets = []
    for band in range(bands):
        if requested_data_min is not None and requested_data_max is not None:
            scale = (requested_data_max - requested_data_min) / (
                scale_data_max - scale_data_min
            )
            offset = requested_data_min - scale * float(scale_data_min)
        else:
            scale = 1.0
            offset = 0.0
        scales += [scale]
        offsets += [offset]
    return scales, offsets


def scale_geotiff_data(
    plot_data, mpl_colors_info, scale_data_min=1, scale_data_max=255, missing_value=0
):
    """Scale geotiff data between scale_data_min and scale_data_max."""
    from geoips.data_manipulations.corrections import apply_data_range

    bands = 1
    if len(plot_data.shape) == 3:
        bands = plot_data.shape[2]

    alpha = None
    if bands == 4:
        # Only scale RGB bands
        bands = 3
        # Pull out the alpha layer, should be 0 or 255. Starts as 0 or 1.
        # 1 + scale*254, for range 1 to 255
        alpha = plot_data[:, :, 3] * 255

    min_val = None
    max_val = None
    inverse = False
    # Use the min and max values in mpl_colors_info norm if specified.
    if (
        mpl_colors_info
        and "norm" in mpl_colors_info
        and hasattr(mpl_colors_info["norm"], "vmax")
    ):
        min_val = mpl_colors_info["norm"].vmin
        max_val = mpl_colors_info["norm"].vmax

    # This is 254 for range 1 to 255.
    num_colors = scale_data_max - scale_data_min

    # Get the scales and offsets for the non-alpha layer.
    scales, offsets = get_scales_and_offsets(
        bands, min_val, max_val, scale_data_min, scale_data_max
    )

    if bands == 1:
        # Just get the scale data with one band
        scale_data = (
            scale_data_min
            + apply_data_range(
                plot_data, min_val=min_val, max_val=max_val, inverse=inverse, norm=True
            )
            * num_colors
        )
    else:
        # This is RGBA, get scaled data for each band
        for band in range(bands):
            curr_scale_data = (
                scale_data_min
                + apply_data_range(
                    plot_data[:, :, band],
                    min_val=min_val,
                    max_val=max_val,
                    inverse=inverse,
                    norm=True,
                )
                * num_colors
            )
            #
            if band == 0:
                scale_data = curr_scale_data
            else:
                scale_data = numpy.dstack((scale_data, curr_scale_data))

    # Add the alpha layer back on.
    if alpha is not None:
        scale_data = numpy.dstack((scale_data, alpha))
        scales += [1.0]
        offsets += [0.0]

    # Set the missing value
    scale_data.fill_value = missing_value
    return scale_data.filled(), missing_value, scales, offsets


def get_geotiff_transform(area_def):
    """Get the appropriate GeoTIFF transform and projectio information."""
    # Note the original implementation here, following the
    # rasterio documentation, resulted in upside down imagery.
    # https://github.com/rasterio/rasterio/issues/1683
    # Updates 20231005 resolved the flipped geotiff imagery.

    # Some notes on geotiff lat/lon ranges:
    # * -180 to 180 longitude range works, anything outside that range will get
    #   cut off (for longlat / 4326)
    # * -90 to 90 latitude range works, but if you go outside that range at all,
    #   the image will not display at all
    # * Affine transform will NOT work if you cross the dateline for 4326,
    #   because of the above
    # * I am not 100% sure if there is a way to define the 4326 projection
    #   to allow crossing the dateline, perhaps if we cross the dateline we
    #   would have to reproject using a different prime meridian?

    # These should be the same as area_extent_ll
    # lons, lats = area_def.get_lonlats()
    # bottom_right_lon = wrap_longitudes(lons[-1][-1])
    # bottom_left_lon = wrap_longitudes(lons[-1][0])
    # top_left_lon = wrap_longitudes(lons[0][0])
    # top_right_lon = wrap_longitudes(lons[0][-1])

    height = area_def.height
    width = area_def.width

    # Note this may need an additional check that the units are in meters for
    # eqc, and units in degrees for longlat projection.  This really requires
    # very specifically set up area definitions that match what cogeotiff is
    # expecting.  I expect we will just add additional try/excepts, and error
    # if the area definition is not of the correct format, rather than adding
    # more supported cases.  We should think carefully about what projections/
    # area definition formats we want to handle / support.
    if area_def.proj_dict["proj"] == "eqc":
        # For eqc projection, area_extent is in m, and area_extent_ll in degrees
        minlat = area_def.area_extent_ll[1]
        maxlat = area_def.area_extent_ll[3]
        minlon = area_def.area_extent_ll[0]
        maxlon = area_def.area_extent_ll[2]
        # Pulling crs from area def did not work for the standard eqc sectors.
        crs = "EPSG:4326"
    elif area_def.proj_dict["proj"] == "longlat":
        # For longlat projection, area_extent is actually in degrees
        minlat = area_def.area_extent[1]
        maxlat = area_def.area_extent[3]
        minlon = area_def.area_extent[0]
        maxlon = area_def.area_extent[2]
        # Pulling crs from area def worked for the longlat projections
        crs = rasterio.crs.CRS.from_proj4(area_def.proj4_string)

    # We may want to add additional checks to ensure sectors are set up properly.
    if maxlon < 0 and minlon > maxlon:
        raise OutputFormatterDatelineError(
            "Region CAN NOT cross the dateline. "
            "Try again with a different sector."
            "Note you can create a projection 'longlat' area_def with the "
            "prime meridian set to the center of the sector with the 'pm' field, "
            "which will allow creating sectors that cross the IDL."
        )

    # Determine the transform based on min/max lat/lon, and width/height.
    res_deg_x = abs(maxlon - minlon) / (width)
    res_deg_y = (maxlat - minlat) / (height)
    src_transform = Affine(res_deg_x, 0, minlon, 0, -res_deg_y, maxlat)

    # Return resulting projection information
    return height, width, crs, src_transform


def call(
    area_def,
    xarray_obj,
    product_name,
    output_fnames,
    product_name_title=None,
    mpl_colors_info=None,
    existing_image=None,
    cog=True,
):
    """Create standard geotiff output using rasterio."""
    # We may want to add additional checks to ensure projections are set up
    # properly for what is explicitly supported in the cogeotiff output formatter,
    # for now it at least requires eqc or longlat, and there must be no lat/lon
    # values outside the -180 to 180 range. There are probably additional
    # requirements for those projections, so we should add to the checks as
    # those requirements are identified.
    if area_def.proj_dict["proj"] not in ["eqc", "longlat"]:
        raise OutputFormatterInvalidProjectionError(
            "Projection must be 'eqc' in order to produce COGS output. "
            "This writer relies on equally spaced lat/lons"
        )

    # Scale the data to produce an 8 bit geotiff.  Defaults to range 1 to 255
    # with missing value of 0.
    plot_data, missing_value, scales, offsets = scale_geotiff_data(
        xarray_obj[product_name].to_masked_array(), mpl_colors_info
    )

    # https://cogeotiff.github.io/rio-cogeo/API/

    with MemoryFile() as memfile:

        # Get the CRS and transform information based on the current area
        # definition.  GeoTIFFs are fairly particular about how the projection
        # information is specified.  Ie, you can't go outside -180 to 180.
        height, width, crs, src_transform = get_geotiff_transform(area_def)

        # If this is RGBA data, e.g. shape is (bands, width, height),
        # then reshape the data as required to write out to geotiff.
        if len(plot_data.shape) == 3 and plot_data.shape[2] == 4:
            is_rgba = True
            bands = plot_data.shape[2]
            plot_data = reshape_as_raster(plot_data.astype(uint8))
            indexes = ()
            for i in range(bands):
                indexes += (i + 1,)
        # Otherwise, we just have a single channel, one band and one index.
        else:
            is_rgba = False
            bands = 1
            indexes = 1

        # Profile for writing the GeoTIFF, using information obtained above.
        src_profile = dict(
            driver="GTiff",
            dtype=uint8,
            count=bands,
            height=height,
            width=width,
            crs=crs,
            transform=src_transform,
        )

        # Initially open the file in memory, based on the above profile.
        with memfile.open(**src_profile) as mem:

            # If this is not RGBA, then set up the internal colormap information.
            if not is_rgba:
                # Populate the input file with numpy array
                cmap_dict = get_rasterio_cmap_dict(mpl_colors_info["cmap"])
                # Must set nodata prior to writing colormap for cog_translate
                mem.nodata = missing_value
                mem.write_colormap(1, cmap_dict)

            # Write out the data to the in-memory GeoTIFF, with the appropriate
            # indexes.
            # masked=True ensures the alpha layer is written to the GeoTIFF,
            # if it is available.  If there is no alpha layer, masked=True
            # has no impact.
            mem.write(plot_data.astype(uint8), indexes=indexes)
            mem.scales = scales
            mem.offsets = offsets

            # This seems to be the default
            # if is_rgba:
            #     mem.colorinterp = [
            #           ColorInterp.red,
            #           ColorInterp.green,
            #           ColorInterp.blue,
            #           ColorInterp.alpha]

            # Adding this explicitly definitely caused more overviews to be
            # created, but I am not sure we want that.  It should come up with
            # reasonable defaults on it's own, and for the very large high
            # resolution sectors we definitely will get more than 3 overviews.
            # if cog:
            #     mem.build_overviews([1, 2, 3])

            dst_profile = cog_profiles.get("deflate")
            for output_fname in output_fnames:
                make_dirs(os.path.dirname(output_fname))
                cog_translate(
                    mem,
                    output_fname,
                    dst_profile,
                    # This should be auto-determined, and I don't think we will
                    # ever want to hard code it here.  Large high resolution
                    # sectors should always end up with more than 3 overviews.
                    # overview_level=3,
                    in_memory=True,
                    quiet=True,
                )
            LOG.info(f"cogeotiff {output_fname}")

    return output_fnames
